/*
 * Created on: Aug 7, 2024
 */

#include <stdint.h>

#include "hsv2rgb.h"
#include "led.h"
#include "rand.h"

#include "ledprog_flames.h"



static uint16_t rnd;
static uint16_t work[4];



/*
 *
 */
static void prog_0_flames(uint8_t tick)
{
    uint8_t i;
    uint16_t j;

    uint16_t hue = 0;   // straight red

    if ((tick & 0x3) == 0) {
        for (i = 0; i < 5; i++) {
            work[0] = 8;
            j = prng_get8();
            if (j > 128) {
                work[0] += j >> 1;
            }

            if (j > 64) {
                // put some orange-green hue in there sometimes
                j = (j >= 0xfd) ? (5*6) : 0;

                hsv2rgb_8b(hue + j, 255, work[0] & 0xff,
                        &led.ind.rgb[i][0], &led.ind.rgb[i][1], &led.ind.rgb[i][2]);
            }
        }

        led_matrix_is_updated();
    }

    // update orange more slowly (why? dunno)
    if ((tick & 0x7) == 3) {
        for (i = 0; i < 3; i++) {
            work[0] = 32;
            j = prng_get8();
            if (j > 0x7f) {
                work[0] += j >> 1;
            }

            led.ind.led[i] = work[0];
        }

        led_matrix_is_updated();
    }
}

/*
 * flames
 */
static void prog_1_rainbow(uint8_t tick)
{
    uint8_t i;
    uint16_t hue;

    if ((tick & 0x3) == 0) {
        work[0] += 1;
        work[0] &= 0xff;
        hue = (255 - work[0]) * 6;

        for (i = 0; i < 5; i++) {
            hsv2rgb_8b(hue, 255, 255, &led.ind.rgb[i][0], &led.ind.rgb[i][1], &led.ind.rgb[i][2]);

            hue += 100;
            hue %= 1536;
        }

        for (i = 0; i < 3; i++) {
            led.ind.led[i] = 0;
        }

        led_matrix_is_updated();
    }
}


static void prog_2_flash(uint8_t tick)
{
    uint8_t i;

    uint16_t hue = 0;
    uint8_t val;

    if (!work[1]) {
        // reset flash timer
        work[1] = 32;
        work[1] += prng_get8() >> 2;

        // do a flash
        work[0] ^= 1;

        if (!work[0]) {
            led.ind.rgb[0][0] = 0;
            led.ind.rgb[0][1] = 0;
            led.ind.rgb[0][2] = 0;

            led.ind.led[0] = led.ind.led[1] = led.ind.led[2] = 0;
        } else {
            hue += prng_get8() >> 4;
            val = (255 - 63) + (prng_get8() >> 2);

            hsv2rgb_8b(hue, 255, val, &led.ind.rgb[0][0], &led.ind.rgb[0][1], &led.ind.rgb[0][2]);

            led.ind.led[0] = led.ind.led[1] = led.ind.led[2] = 255;
        }

        for (i = 1; i < 5; i++) {
            led.ind.rgb[i][0] = led.ind.rgb[0][0];
            led.ind.rgb[i][1] = led.ind.rgb[0][1];
            led.ind.rgb[i][2] = led.ind.rgb[0][2];
        }

        led_matrix_is_updated();
    }

    work[1]--;
}

/*
 * iterate over previous programs after random delays
 */
static uint8_t iterate_tmr;
static uint8_t iterate_prg;
static void prog_3_iterate(uint8_t tick)
{
    uint8_t w[2];

    if (!iterate_tmr) {
        // between 26 - 90 seconds per program
        iterate_tmr = prng_get8() >> 2;
        iterate_tmr += 26;

        iterate_prg++;

        // this resets our counters so save then restore them
        w[0] = iterate_tmr;
        w[1] = iterate_prg;
        ledprog_top_init();
        iterate_tmr = w[0];
        iterate_prg = w[1];
    }

    if (!tick) {
        iterate_tmr--;
    }

    // run all but the last two programs (this program, and off program)
    if (iterate_prg > (sizeof(ledprog_flames) / 4) - 2) {
        iterate_prg = 1;
    }

    if (ledprog_flames[iterate_prg - 1]) {
        ledprog_flames[iterate_prg - 1](tick);
    }
}

static void prog_4_off(uint8_t tick)
{
    uint8_t i;

    // blank it
    for (i = 0; i < sizeof(led.all); i++) {
        led.all[i] = 0;
    }

    if (!work[0]) {
        led_matrix_is_updated();
        work[0] = 1;
    }
}



void (*ledprog_flames[5])(uint8_t) = {
    prog_0_flames,
    prog_1_rainbow,
    prog_2_flash,
    prog_3_iterate,
    prog_4_off
};



void ledprog_top_init()
{
    uint8_t i;

    rnd = prng_get16();

    // global program initialization
    for (i = 0; i < 4; i++) {
        work[i] = 0;
    }

    // program specific init
    iterate_tmr = 0;
    iterate_prg = 0;
}
